__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2019 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging

from octoprint.util import monotonic_time


class Check:
    name = None
    url = None

    ACTIVE_TIMEOUT = 30.0

    def __init__(self):
        self._active = True
        self._triggered = False
        self._start_time = None
        self.placeholders = {}

    def received(self, line):
        """Called when receiving a new line from the printer"""
        pass

    def m115(self, name, data):
        """Called when receiving the response to an M115 from the printer"""
        pass

    def cap(self, cap, enabled):
        """Called when receiving a capability report line"""
        pass

    @property
    def active(self):
        """Whether this check is still active"""
        return self._active

    @property
    def triggered(self):
        """Whether the check has been triggered"""
        return self._triggered

    def evaluate_timeout(self):
        if self._start_time is None:
            self._start_time = monotonic_time()
        else:
            if monotonic_time() > self._start_time + self.ACTIVE_TIMEOUT:
                # timeout ran out, this is now inactive
                self._active = False
                logging.getLogger(__name__).info(
                    "Deactivated {} due to timeout after {}s".format(
                        __name__, self.ACTIVE_TIMEOUT
                    )
                )

    def reset(self):
        self._active = True
        self._triggered = False
        self._start_time = None


class LineCheck(Check):
    def __init__(self):
        Check.__init__(self)
        self._matches = None

    def received(self, line):
        if not line:
            return

        if self._is_match(line):
            self._matches = True
        elif self._is_ruled_out(line):
            self._matches = False

        self._evaluate()

    def m115(self, name, data):
        # M115 usually means we stop scanning received lines
        if self._matches is None:
            self._matches = False
        self._evaluate()

    def _evaluate(self):
        if self._matches is None:
            return
        self._triggered = self._matches
        self._active = False

    def reset(self):
        Check.reset(self)
        self._matches = None

    def _is_match(self, line):
        return False

    def _is_ruled_out(self, line):
        return False


class CapCheck(Check):
    CAP = None

    def __init__(self):
        Check.__init__(self)
        self.cap_seen = False

    def cap(self, cap, enabled):
        self.cap_seen = True
        if cap == self.CAP:
            self._triggered = self._eval(enabled)
            self._active = False

    def received(self, line):
        if self.cap_seen and not line.startswith("Cap:"):
            # first non cap line after cap report: cap report over, deactivate
            self._active = False

    def _eval(self, enabled):
        # trigger when enabled
        return enabled


class NegativeCapCheck(CapCheck):
    def _eval(self, enabled):
        # trigger when NOT enabled
        return not enabled


class AuthorCheck(Check):
    authors = ()

    AUTHOR = "| Author: ".lower()

    def m115(self, name, data):
        # M115 usually means we stop scanning received lines
        self._active = False

    def received(self, line):
        if not line:
            return

        lower_line = line.lower()
        if self.AUTHOR in lower_line:
            self._triggered = any(x in lower_line for x in self.authors)
            self._active = False


class Severity:
    INFO = "info"
    WARNING = "warning"
    CRITICAL = "critical"



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"

import textwrap

import flask
import octoprint.plugin
from flask_babel import gettext
from octoprint.access import USER_GROUP
from octoprint.access.permissions import Permissions
from octoprint.events import Events
from octoprint.util import to_unicode
from octoprint.util.version import is_octoprint_compatible

from .checks import Severity
from .checks.firmware_broken import FirmwareBrokenChecks
from .checks.firmware_development import FirmwareDevelopmentChecks
from .checks.firmware_hostcommands import FirmwareHostcommandsChecks
from .checks.firmware_unsafe import FirmwareUnsafeChecks

TERMINAL_WARNING = """
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
{message}

Learn more at {url}
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

"""

TERMINAL_INFO = """
---------------------------------------------------------------------------
{message}

Learn more at {url}
---------------------------------------------------------------------------

"""

FIRMWARE_CHECKS = {
    "firmware-unsafe": FirmwareUnsafeChecks.as_dict(),
    "firmware-broken": FirmwareBrokenChecks.as_dict(),
    "firmware-development": FirmwareDevelopmentChecks.as_dict(),
    "firmware-hostcommands": FirmwareHostcommandsChecks.as_dict(),
}


class FirmwareCheckPlugin(
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.EventHandlerPlugin,
    octoprint.plugin.SimpleApiPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.SettingsPlugin,
):
    # noinspection PyMissingConstructor
    def __init__(self):
        self._warnings = {}
        self._scan_received = True

    ##~~ TemplatePlugin API

    def get_template_configs(self):
        return [
            {
                "type": "sidebar",
                "name": gettext("Attention!"),
                "data_bind": "visible: printerState.isOperational() && loginState.isAdmin() && warnings().length > 0",
                "icon": "exclamation-triangle",
                "styles_wrapper": ["display: none"],
                "template": "firmware_check_sidebar_warning.jinja2",
                "suffix": "_warning",
            },
            {
                "type": "sidebar",
                "name": gettext("Info"),
                "data_bind": "visible: printerState.isOperational() && loginState.isAdmin() && infos().length > 0",
                "icon": "info-circle",
                "styles_wrapper": ["display: none"],
                "template": "firmware_check_sidebar_info.jinja2",
                "suffix": "_info",
            },
            {
                "type": "settings",
                "name": gettext("Firmware Check"),
                "custom_bindings": False,
            },
        ]

    ##~~ AssetPlugin API

    def get_assets(self):
        return {
            "js": ("js/firmware_check.js",),
            "clientjs": ("clientjs/firmware_check.js",),
            "css": ("css/firmware_check.css",),
            "less": ("less/firmware_check.less",),
        }

    ##~~ EventHandlerPlugin API

    def on_event(self, event, payload):
        if event == Events.DISCONNECTED:
            self._reset_warnings()
            self._reset_state()
            self._reset_checks()

    ##~~ SimpleApiPlugin API

    def on_api_get(self, request):
        if not Permissions.PLUGIN_FIRMWARE_CHECK_DISPLAY.can():
            return flask.make_response("Insufficient rights", 403)
        return flask.jsonify(self._warnings)

    def is_api_protected(self):
        return True

    ##~~ SettingsPlugin API

    def get_settings_defaults(self):
        return {"ignore_infos": False}

    ##~~ GCODE received hook handler

    def on_gcode_received(self, comm_instance, line, *args, **kwargs):
        if self._scan_received:
            self._run_checks("received", to_unicode(line, errors="replace"))
        return line

    ##~~ Firmware info hook handler

    def on_firmware_info_received(self, comm_instance, firmware_name, firmware_data):
        self._run_checks(
            "m115",
            to_unicode(firmware_name, errors="replace"),
            {
                to_unicode(key, errors="replace"): to_unicode(value, errors="replace")
                for key, value in firmware_data.items()
            },
        )

    ##~~ Firmware capability hook handler

    def on_firmware_cap_received(self, comm_instance, cap, enabled, all_caps):
        self._run_checks("cap", to_unicode(cap, errors="replace"), enabled)

    ##~~ Additional permissions hook handler

    def get_additional_permissions(self):
        return [
            {
                "key": "DISPLAY",
                "name": "Display firmware check warnings",
                "description": gettext("Allows to see firmware check warnings"),
                "roles": ["display"],
                "default_groups": [USER_GROUP],
            }
        ]

    ##~~ Softwareupdate hook

    def get_update_information(self):
        return {
            "firmware_check": {
                "displayName": "Firmware Check Plugin",
                "displayVersion": self._plugin_version,
                # version check: github repository
                "type": "github_release",
                "user": "OctoPrint",
                "repo": "OctoPrint-FirmwareCheck",
                "current": self._plugin_version,
                "stable_branch": {
                    "name": "Stable",
                    "branch": "master",
                    "commitish": ["devel", "master"],
                },
                "prerelease_branches": [
                    {
                        "name": "Prerelease",
                        "branch": "devel",
                        "commitish": ["devel", "master"],
                    }
                ],
                # update method: pip
                "pip": "https://github.com/OctoPrint/OctoPrint-FirmwareCheck/archive/{target_version}.zip",
            }
        }

    ##~~ Helpers

    def _run_checks(self, check_type, *args, **kwargs):
        changes = False

        still_active = False
        for warning_type, check_data in FIRMWARE_CHECKS.items():
            checks = check_data.get("checks")
            message = check_data.get("message")
            severity = check_data.get("severity", Severity.CRITICAL)
            url = "https://faq.octoprint.org/warning-{warning_type}".format(
                warning_type=warning_type
            )
            if not checks or not message:
                continue

            for check in checks:
                if not check.active:
                    # skip non active checks
                    continue

                method = getattr(check, check_type, None)
                if not callable(method):
                    # skip uncallable checks
                    continue

                # execute method
                try:
                    method(*args, **kwargs)
                except Exception:
                    self._logger.exception(
                        "There was an error running method {} on check {!r}".format(
                            check_type, check
                        )
                    )
                    continue

                # check if now triggered
                if check.triggered:
                    if check.url is not None:
                        url = check.url

                    self._register_warning(
                        warning_type,
                        message.format(**check.placeholders),
                        severity,
                        url,
                    )

                    # noinspection PyUnresolvedReferences
                    self._event_bus.fire(
                        Events.PLUGIN_FIRMWARE_CHECK_WARNING,
                        {
                            "check_name": check.name,
                            "warning_type": warning_type,
                            "severity": severity,
                            "url": url,
                        },
                    )
                    changes = True
                    break

                check.evaluate_timeout()
                self._logger.debug(f"Check {check} active? {check.active}")

            still_active = any(check.active for check in checks) or still_active

        self._scan_received = still_active
        self._logger.debug(f"Scanning received lines enabled? {self._scan_received}")

        if changes:
            self._ping_clients()

    def _register_warning(self, warning_type, message, severity, url):
        output = TERMINAL_WARNING
        if severity == Severity.INFO:
            output = TERMINAL_INFO

        self._log_to_terminal(
            output.format(
                message="\n".join(textwrap.wrap(message, 75)),
                warning_type=warning_type,
                url=url,
            )
        )
        self._warnings[warning_type] = {
            "message": message,
            "severity": severity,
            "url": url,
        }

        logline = f"{message}. More information at {url}"
        if severity == Severity.INFO:
            self._logger.info(logline)
        else:
            self._logger.warning(logline)

    def _reset_warnings(self):
        self._warnings.clear()
        self._ping_clients()

    def _reset_state(self):
        self._scan_received = True

    def _reset_checks(self):
        for _, check_data in FIRMWARE_CHECKS.items():
            checks = check_data.get("checks")
            if not checks:
                continue

            for check in checks:
                check.reset()

    def _log_to_terminal(self, message):
        if self._printer:
            lines = message.split("\n")
            self._printer.log_lines(*lines)

    def _ping_clients(self):
        self._plugin_manager.send_plugin_message(self._identifier, {"type": "update"})


def register_custom_events(*args, **kwargs):
    return [
        "warning",
    ]


__plugin_name__ = "Firmware Check"
__plugin_pythoncompat__ = ">=3.7,<4"
__plugin_disabling_discouraged__ = gettext(
    "Without this plugin OctoPrint will no longer be able to "
    "check if the printer it is connected to has a known safety "
    "issue or otherwise broken firmware and inform you about that fact."
)
__plugin_implementation__ = FirmwareCheckPlugin()
__plugin_hooks__ = {
    "octoprint.plugin.softwareupdate.check_config": __plugin_implementation__.get_update_information,
    "octoprint.comm.protocol.gcode.received": (
        __plugin_implementation__.on_gcode_received,
        100,
    ),
    "octoprint.comm.protocol.firmware.info": (
        __plugin_implementation__.on_firmware_info_received,
        100,
    ),
    "octoprint.comm.protocol.firmware.capabilities": (
        __plugin_implementation__.on_firmware_cap_received,
        100,
    ),
    "octoprint.events.register_custom_events": register_custom_events,
    "octoprint.access.permissions": __plugin_implementation__.get_additional_permissions,
}

if is_octoprint_compatible("<1.6"):
    __plugin_settings_overlay__ = {
        "appearance": {
            "components": {
                "order": {
                    "sidebar": [
                        "plugin_firmware_check_warning",
                        "plugin_firmware_check_info",
                        "connection",
                        "state",
                        "files",
                    ]
                }
            }
        }
    }



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2021 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask_babel import gettext

from . import NegativeCapCheck, Severity


class FirmwareHostcommandsChecks:
    @classmethod
    def as_dict(cls):
        return {
            "checks": (FirmwareHostcommandsCapCheck(),),
            "message": gettext(
                "Your printer's firmware supports host action commands, but they are "
                "disabled. The firmware will not inform OctoPrint about job pause & "
                "cancellations, filament runouts, and similar firmware-side events. If "
                "you want this functionality, you need to change your firmware's "
                "configuration."
            ),
            "severity": Severity.INFO,
        }


class FirmwareHostcommandsCapCheck(NegativeCapCheck):
    """
    Firmware reporting disabled HOST_ACTION_COMMANDS capability
    """

    name = "capability"
    CAP = "HOST_ACTION_COMMANDS"



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2020 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask_babel import gettext
from octoprint.settings import settings

from . import LineCheck, Severity


class FirmwareBrokenChecks:
    @classmethod
    def as_dict(cls):
        return {
            "checks": (
                CbdCheck(),
                ZwlfCheck(),
                CrealityDoubleTempCheck(),
                CrealityTFCardCheck(),
            ),
            "message": gettext(
                "Your printer's firmware is known to have a broken implementation of the "
                "communication protocol. This may cause print failures or other annoyances. "
                "You'll need to take additional steps for OctoPrint to fully work with it."
            ),
            "severity": Severity.WARNING,
        }


class CbdCheck(LineCheck):
    name = "cbd"
    url = "https://faq.octoprint.org/warning-firmware-broken-cbd"
    FRAGMENT = "CBD make it".lower()

    def _is_match(self, line):
        return self.FRAGMENT in line.lower()


class ZwlfCheck(CbdCheck):
    name = "zwlf"
    FRAGMENT = "ZWLF make it".lower()


class CrealityDoubleTempCheck(LineCheck):
    name = "creality_double_temp"
    url = "https://faq.octoprint.org/warning-firmware-broken-creality-double-temp"

    def m115(self, name, data):
        # we cannot stop scanning after an M115 report as we might not yet have seen a temperature report then
        pass

    def _is_match(self, line):
        # broken report: TT::27.9327.93 //0.000.00 BB::39.6739.67 //0.000.00 @@::00 BB@@::00
        return "TT::" in line

    def _is_ruled_out(self, line):
        # first thing that looks like a proper report stops scanning
        return (
            " T:" in line
            or line.startswith("T:")
            or " T0:" in line
            or line.startswith("T0:")
            or ((" B:" in line or line.startswith("B:")) and "A:" not in line)
        )


class CrealityEqualsTempCheck(CrealityDoubleTempCheck):
    name = "creality_equals_temp"
    url = "https://faq.octoprint.org/warning-firmware-broken-creality-equals-temp"

    def _is_match(self, line):
        # broken report: ==T:145.66 /210.00 ==B:60.15 /60.00 @:127 B@:0
        return "==T:" in line


class CrealityTFCardCheck(LineCheck):
    name = "creality_tfcard"
    url = "https://faq.octoprint.org/warning-firmware-broken-creality-tfcard"

    def m115(self, name, data):
        # we cannot stop scanning after an M115 report as we might not yet have seen an SD card report then
        pass

    def _is_match(self, line):
        return "TF card ok" in line

    def _is_ruled_out(self, line):
        # first thing that looks like a proper report stops scanning
        lower_line = line.lower()
        return not settings().getBoolean(["feature", "sdSupport"]) or any(
            x in lower_line
            for x in ("sd card ok", "sd init fail", "sd printing byte", "not sd printing")
        )



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2019 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask_babel import gettext
from octoprint.util.version import get_comparable_version

from . import AuthorCheck, Check, NegativeCapCheck, Severity


class FirmwareUnsafeChecks:
    @classmethod
    def as_dict(cls):
        return {
            "checks": (
                AnetA8Check(),
                AnycubicCheck(),
                CrealityCR10sCheck(),
                CrealityEnder3Check(),
                FlashforgeVoxelabAquilaCheck(),
                MalyanM200Check(),
                Micro3DIMECheck(),
                Micro3DStockCheck(),
                RepetierBefore092Check(),
                ThermalProtectionCapCheck(),
            ),
            "message": gettext(
                "Your printer's firmware is known to lack mandatory safety features (e.g. "
                "thermal runaway protection). This is a fire risk."
            ),
            "severity": Severity.CRITICAL,
        }


class AnetA8Check(Check):
    """
    Anet A8 stock firmware

    Identified through firmware name "ANET_A8_".
    """

    name = "anet_a8"

    def m115(self, name, data):
        self._triggered = name and name.lower().startswith("anet_a8_")
        self._active = False


class AnycubicCheck(Check):
    """
    Anycubic MEGA stock firmware

    Identified through "Author: (Jolly, xxxxxxxx.CO.)" or "| Author: (**Jolly, xxxxxxxx.CO.**)" in startup messages
    combined with "echo:Vx.y.z" in startup messages, with x.y.z < 1.1.2.
    """

    name = "anycubic"

    AUTHOR = "| Author: ".lower()
    VERSION = "echo:V"

    CRITICAL_AUTHOR1 = "| Author: (Jolly, xxxxxxxx.CO.)".lower()
    CRITICAL_AUTHOR2 = "| Author: (**Jolly, xxxxxxxx.CO.**)".lower()

    FIXED_VERSION = get_comparable_version("1.1.2")

    def __init__(self):
        Check.__init__(self)
        self._author_matches = None
        self._version_matches = None

    def m115(self, name, data):
        # make sure we stop scanning once we see the response to M115
        if self._author_matches is None:
            self._author_matches = False
        if self._version_matches is None:
            self._version_matches = False
        self._evaluate()

    def received(self, line):
        if not line:
            return

        lower_line = line.lower()
        if self.AUTHOR in lower_line:
            self._author_matches = (
                self.CRITICAL_AUTHOR1 in lower_line or self.CRITICAL_AUTHOR2 in lower_line
            )
        elif line.startswith(self.VERSION):
            self._version_matches = self._broken_version(line)
        else:
            return

        self._evaluate()

    def _broken_version(self, line):
        version_str = line[len(self.VERSION) :]
        try:
            version = get_comparable_version(version_str, base=True)
        except Exception:
            version = None

        if version is not None and version < self.FIXED_VERSION:
            return True
        else:
            return False

    def _evaluate(self):
        if self._author_matches is None or self._version_matches is None:
            return
        self._triggered = self._author_matches and self._version_matches
        self._active = False

    def reset(self):
        Check.reset(self)
        self._author_matches = None
        self._version_matches = None


class CrealityCR10sCheck(AuthorCheck):
    """
    Creality CR10s

    Identified through " | Author: (CR-10Slanguage)" in startup messages.
    """

    name = "creality_cr10s"
    authors = (" | Author: (CR-10Slanguage)".lower(),)


class CrealityEnder3Check(AuthorCheck):
    """
    Creality Ender3

    Identified through " | Author: (Ender3)" in startup messages.
    """

    name = "creality_ender3"
    authors = (" | Author: (Ender3)".lower(),)


class FlashforgeVoxelabAquilaCheck(Check):
    """
    Flashforge Voxelab Aquila stock firmware

    Identified through "MACHINE_TYPE:Voxlab-1 V1" (sic!) in M115 response.
    """

    name = "flashforge_voxelab_aquila"
    MACHINE_TYPE = "Voxlab-1 V1"  # sic!

    def m115(self, name, data):
        self._triggered = data.get("MACHINE_TYPE") == self.MACHINE_TYPE
        self._active = False


class MalyanM200Check(Check):
    """
    Malyan M200 stock firmware prior to version 4.0

    Identified through firmware name "Malyan*", model "M200" and version < 4.0.
    """

    name = "malyan_m200"

    FIXED_VERSION = get_comparable_version("4.0")

    def m115(self, name, data):
        try:
            version = get_comparable_version(data.get("VER", "0"))
        except Exception:
            version = None

        self._triggered = (
            name
            and name.lower().startswith("malyan")
            and data.get("MODEL") == "M200"
            and version is not None
            and version < self.FIXED_VERSION
        )
        self._active = False


class Micro3DIMECheck(Check):
    """
    Micro3D with IME firmware

    Identified through firmware name "iME*".
    """

    name = "micro3d_ime"

    def m115(self, name, data):
        self._triggered = name and name.lower().startswith("ime")
        self._active = False


class Micro3DStockCheck(Check):
    """
    Micro3D with IME firmware

    Identified through firmware name "Micro3D*".
    """

    name = "micro3d"

    def m115(self, name, data):
        self._triggered = name and name.lower().startswith("micro3d")
        self._active = False


class RepetierBefore092Check(Check):
    """
    Repetier firmware prior to version 0.92

    Identified through firmware name "Repetier_x.y.z" with x.y.z < 0.92
    """

    name = "repetier_before_092"

    FIXED_VERSION = get_comparable_version("0.92")

    def m115(self, name, data):
        if name and name.lower().startswith("repetier"):
            version = self._extract_repetier_version(name)
            self._triggered = version is not None and version < self.FIXED_VERSION
        self._active = False

    def _extract_repetier_version(self, name):
        """
        Extracts the Repetier version number from the firmware name.

        Example: "Repetier_0.91" => 0.91
        """
        version = None
        if "_" in name:
            _, version = name.split("_", 1)
            try:
                version = get_comparable_version(version, base=True)
            except Exception:
                pass
        return version


class ThermalProtectionCapCheck(NegativeCapCheck):
    """
    Firmware reporting disabled THERMAL_PROTECTION capability
    """

    name = "capability"
    CAP = "THERMAL_PROTECTION"



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2019 The OctoPrint Project - Released under terms of the AGPLv3 License"

import datetime
import re

from flask_babel import gettext

from . import Check, Severity


class FirmwareDevelopmentChecks:
    @classmethod
    def as_dict(cls):
        return {
            "checks": (
                MarlinBugfixCheck(),
                MarlinMfsBugfixCheck(),
            ),
            "message": gettext(
                "Your printer's firmware is a {buildtype} build of {firmware} "
                "(build date {builddate}). It might be more unstable "
                "than a release version and should be kept up-to-date."
            ),
            "severity": Severity.INFO,
        }


class MarlinBugfixCheck(Check):
    """
    Marlin bugfix builds.

    Identified by firmware name that contains "Marlin bugfix-<version> (<builddate>)"
    """

    name = "marlin_bugfix"

    pattern = re.compile(
        r"marlin (?P<version>bugfix-[^-]*)\s\((?P<builddate>.*)\)",
    )

    def __init__(self):
        Check.__init__(self)
        self.placeholders = {
            "firmware": "Marlin",
            "version": "unknown",
            "builddate": "unknown",
            "buildtype": "development",
        }

    def m115(self, name, data):
        match = self.pattern.match(name.lower())
        if match:
            self.placeholders["version"] = match.group("version")
            self.placeholders["builddate"] = match.group("builddate")

            try:
                self.placeholders["builddate"] = datetime.datetime.strptime(
                    match.group("builddate"), "%b %d %Y %H:%M:%S"
                ).strftime("%Y%m%d")
            except Exception:
                pass

        self._triggered = name and match
        self._active = False


class MarlinMfsBugfixCheck(MarlinBugfixCheck):
    """
    Marlin MFS bugfix builds.

    Identified by firmware name that matches "Marlin bugfix-<version>-mfs (<builddate>)"
    """

    name = "marlin_bugfix_mfs"
    url = "https://faq.octoprint.org/warning-firmware-development-mfs"

    pattern = re.compile(
        r"marlin (?P<version>bugfix-.*-mfs)\s\((?P<builddate>.*)\)",
    )

    def __init__(self):
        MarlinBugfixCheck.__init__(self)
        self.placeholders["firmware"] = "Marlin by the Marlin Firmware Service"
        self.placeholders["buildtype"] = "nightly"
